;*****************************************************************************
;*****************************************************************************
;**                                                                         **
;**                      Z8 BATTERY CHARGING SOFTWARE                       **
;**                   CCP Emulator Version rel 12 Dec 94                    **
;**   This software is based on a program and proprietary concepts which    **
;**   were developed by Black & Decker.  Use of this software requires a    **
;**   license agreement.  Changes to the original B&D 2K code:              **
;**   1. Invert current handler subroutines. Chg current ON = logic 1 out.  **
;**   2. The application specific threshold values grouped in a single      **
;**      section of the source code to simplify programming, and values     **
;**      differ from original Black & Decker values which assume a specific **
;**      NiCd cell type.                                                    **
;**   3. Added WDTs to solve spurious WDT resets.  Init IMR to mask-off t0. **
;**   4. NiCd fast full charge detect test is not used as this assumes a    **
;**      specific NiCd cell type.  Routine is included in source code, but  ** 
;**      is ignored by assembler due to leading ";" at each line (REM-out). **
;**   5. Deleted heatsink overtemp sensor test and handling routine.        **
;**   6. Deleted factory test routine specific to B&D manufacturing process **
;**   7. Added conditional assembly features to permit software development **
;**      using low cost Zilog CCP emulator board.  Described below.         **
;**                                                                         **
;*****************************************************************************

;*****************************************************************************
;**                    Conditional Assembly Features                        **
;**                                                                         **
;**     The CCP emulator differs from the Z86B07 device: The 07 does not    **
;**     have autolatches, which cause problems with the battery A/D scheme. **
;**     Contact Zilog for specific emulation hardware circuit compatible    **
;**     with this software.  In addition, the B07 has analog comparator P32 **
;**     rising edge interrupt (int3) used in this application.  The CCP     **
;**     emulator uses int0 to emulate this feature.  See the Z86C30 spec    **
;**     for details of the CCP identical interrupt scheme. Changes for CCP  **
;**     emulation are automatically made by conditional assembly.           **
;**                                                                         **
;**           TWO VERY IMPORTANT NOTES REGARDING CCP EMULATOR USE :         **
;**    1. The log A/D method is very sensitive to VCC noise.  The Z8's      **
;**    VCC supply MUST be well-regulated and free of AC line ripple.        **
;**    When running from the CCP emulator, the Z8 supply is the CCP board's **
;**    power supply, NOT the 5V supply on your prototype charger board.     ** 
;**    BE CERTAIN THAT THE CCP SUPPLY IS FREE FROM RIPPLE, especially if    **
;**    your CCP board is running from an AC adapter (NOT RECOMMENDED for    **
;**    this application.)  This is a likely source of spurious A/D jumps    **
;**    or seemingly random A/D fluctuations which are coincidental with AC  **
;**    line variations or switching of external high current AC loads, like **
;**    AC motors.  Check the DC supply on the CCP if A/D values jump!       **
;**    In your final product design, zener regulation of VCC is adequate,   **
;**              but 100/120 Hz ripple on VCC is not tolerable.             **
;**                                                                         **
;**    2.  For CCP emulator, the following line should read:   CCP .equ 1   **
;**        For masked Z86B07, masked Z86C07 or OTP Z86E07,                  **
;**        edit the source code so the following line reads:   CCP .equ 0   **
;**                                                                         **
;*****************************************************************************

CCP              .equ      1           ; "1" for CCP emulator, otherwise "0"
                  
;P0              .equ      00H         ;PORT 0 INPUTS USED FOR STRAP OPTIONS
                                       ;  BIT 2 SERIAL DATA HI-Z AND LOW.
                                       ;  BIT 1 AND BIT 0 SELECT FAST FULL CHG
                                       ;    DETECT (FFCD) THRESHOLDS.

;P2              .equ      02H         ;I/O PORT 2
                                       ;  BIT 7 UNUSED INPUT
                                       ;  BIT 6 PIEZO OUTOUT
                                       ;  BIT 5 CHGON\ OUTPUT
                                       ;  BIT 4 AD ARM/CONVERT INPUT/OUTPUT
                                       ;  BIT 3 LED OUTPUT; TEST MODE INPUT
                                       ;  BIT 2 PRESCALER INPUT/OUTPUT
                                       ;  BIT 1 PRESCALER INPUT/OUTPUT
                                       ;  BIT 0 PRESCALER INPUT/OUTPUT

;P3              .equ      03H         ;I/O port 3
                                       ;  BIT 3 A/D RAMP INPUT
                                       ;  BIT 2 BATTERY A/D INPUT
                                       ;  BIT 1  SPARE INPUT 

TIMEOUT          .equ      04H         ;TOTAL NUMBER OF SECONDS CHG CURRENT
TIMEOUT_MSB      .equ      04H         ; ALLOWED FOR A GIVEN PACK INSERTION.
TIMEOUT_LSB      .equ      05H

USER_FLAGS       .equ      06H          ;BIT FLAGS.
CURRENT_STATE    .equ      10000000B    ; 1=CURRENT IS ON, 0=CURRENT IS OFF.
HI_Z_DETECTED    .equ      01000000B    ; 1=HIGH IMPED., 0=NOT HIGH IMPED.
IGNORE_REMOVAL   .equ      00100000B    ; 1=IGNORE BATTERY REMOVAL (FOR
                                        ;  AUTOSCALE ROUTINE).

TEMP             .equ      07H          ;GENERAL PURPOSE BYTE STORAGE.

TEMP_2           .equ      08H          ;GENERAL PURPOSE 2 BYTE STORAGE.
TEMP_MSB         .equ      08H
TEMP_LSB         .equ      09H

LOOP_COUNTER     .equ      0AH          ;GENERAL PURPOSE LOOP_COUNTER.

P2M_IMAGE        .equ      0BH          ;MIRRORS PORT 2 MODE FOR READ QUERIES
                                        ; SINCE P2M IS WRITE ONLY. 

SERIAL_OUTPUT_BYTE .equ    0CH          ;THIS BYTE IS SERIAL OUTPUT STORAGE.

TEST_MODE_COUNTER  .equ    0DH          ;USED IN TEST_MODE ROUTINE.

NOTHING_1          .equ    0EH          ;2 PLACE HOLDERS TO GET AD_MSB_0
NOTHING_2          .equ    0FH          ; AND AD_MSB_F TO CORRECT ADDRESS.


;  LOCATIONS 10H THROUGH 1FH GENERALLY SHOULD NOT BE USED AS THE GET_AD 
;  SUBROUTINE USES THESE LOCATIONS. IF THESE REGISTERS ARE USED, A CALL TO 
;  THE GET_AD SUBROUTINE WILL TRASH CONTENTS!

AD_MSB_0         .equ      10H
AD_MSB_F         .equ      1FH

;LOCATIONS 20H THROUGH 2EH ARE PROGRAM STORAGE LOCATIONS, AND CONTAIN VALUES
;THAT ARE TRANSMITTED BY THE SERIAL DATA TRANSMISSION ROUTINES.

MODE             .equ      20H    ;ALPHA CHARACTER IDENTIFIES OPERATING MODE.

CONFIGURATION    .equ      21H    ;BITS 2,1,0 OF PORT_0.

TIME             .equ      22H    ;SYSTEM TIME IN SECONDS.
TIME_MSB         .equ      22H
TIME_LSB         .equ      23H

AD               .equ      24H     ;results of most recent AD conversion.
AD_MSB           .equ      24H
AD_LSB           .equ      25H

SLOPE            .equ      26H     ;results of most recent slope calculations.
SLOPE_MSB        .equ      26H
SLOPE_LSB        .equ      27H

EVEN_ADDR_28     .equ      28H     ;JUNK maintains even address boundries.
INFLECTION       .equ      29H     ;storage for inflection, termination,
                                   ;  and problem codes.  

MSLOPE           .equ      2AH     ;storage for maximum and minimum slopes.
MSLOPE_MSB       .equ      2AH
MSLOPE_LSB       .equ      2BH

MAXAD            .equ      2CH    ;THE MAXIMUM FILTERED AD VALUE FOR F MINUS
MAXAD_MSB        .equ      2CH    ; CALCULATIONS.
MAXAD_LSB        .equ      2DH

ATTENUATOR       .equ      2EH    ;ACTIVE ATTENUATION DIVIDER VALUE.

PROBLEM_MODE     .equ      2FH    ;IDENTIFICATION OF PROBLEM MODE.

                      
;THE AD STACK. THIS IS AN 8 LEVEL, FIFO STACK WHERE AD0 IS MOST RECENT VALUE 
;AND AD7 IS OLDEST. AD READINGS ARE 12 BIT. USED TO GENERATE SLOPE DATA
;BY THE TERMINATION ROUTINES. UPDATED EVERY 16 SECONDS FOR 15 MINUTE CHARGER.

AD0              .equ     30H    ;NEWEST AD READING.
AD0MSB           .equ     30H
AD0LSB           .equ     31H

AD1              .equ     32H
AD1MSB           .equ     32H
AD1LSB           .equ     33H

AD2              .equ     34H
AD2MSB           .equ     34H
AD2LSB           .equ     35H

AD3              .equ     36H
AD3MSB           .equ     36H
AD3LSB           .equ     37H

AD4              .equ     38H
AD4MSB           .equ     38H
AD4LSB           .equ     39H

AD5              .equ     3AH
AD5MSB           .equ     3AH
AD5LSB           .equ     3BH

AD6              .equ     3CH
AD6MSB           .equ     3CH
AD6LSB           .equ     3DH

AD7              .equ     3EH    ;OLDEST AD READING.
AD7MSB           .equ     3EH
AD7LSB           .equ     3FH



; TWO LEVEL FIFO STACK USED BY THE MINUS DELTA VOLTAGE ROUTINE.
; AD0_MINUS_DELTA_V IS NEWEST ENTRY WHILE AD1_MINUS_DELTA_V IS OLDEST. 
; THE STACK IS UPDATED EVERY 4 SECONDS IN 15 MINUTE CHARGER.
; RATE SHOULD BE ADJUSTED FOR 15 SEC UPDATES FOR 1 HOUR CHARGER.

AD0_MINUS_DELTA_V .equ     40H
AD0_MDV_MSB       .equ     40H
AD0_MDV_LSB       .equ     41H

AD1_MINUS_DELTA_V .equ     42H
AD1_MDV_MSB       .equ     42H
AD1_MDV_LSB       .equ     43H



DIVISOR          .equ     44H       ;USED BY THE "DIVIDE" ROUTINE.
DIVISOR_MSB      .equ     44H
DIVISOR_LSB      .equ     45H

Z_TEST_I_ON      .equ     46H       ;HIGH IMPEDENCE TEST CURRENT ON FACTOR 
Z_TEST_I_ON_MSB  .equ     46H       ; STORES THE DERIVED HIGH IMPEDANCE VALUE
Z_TEST_I_ON_LSB  .equ     47H       ; FROM INSERTED PACK.

FFCD_FACTOR       .equ     48H       ;FULL CHARGE DETECTION FACTOR STORES THE
FFCD_FACTOR_MSB   .equ     48H       ; DERIVED FULL CHARGE DETECTION FACTOR
FFCD_FACTOR_LSB   .equ     49H       ; FROM INSERTED PACK.

FFCD_THRESHOLD      .equ   4AH       ;STORAGE FOR FULL CHARGE DETECTION
FFCD_THRESHOLD_MSB  .equ   4AH       ; THRESHOLD WHICH IS SELECTED BY BITS 0
FFCD_THRESHOLD_LSB  .equ   4BH       ; AND 1 OF PORT 0.

;*****************************************************************************
;**       Following EQUs are not required for Zilog Z8 assembler            **
;**           but may be necessary for third party assemblers               **
;*****************************************************************************
                      
;TMR              .equ      F1H          ;timer mode
;T1               .equ      F2H          ;TIMER/COUNTER 1
;PRE1             .equ      F3H          ;TIMER PRESCALER 1

;T0               .equ      F4H          ;TIMER/COUNTER 0
;PRE0             .equ      F5H          ;TIMER PRESCALER 0

;P2M              .equ      F6H          ;PORT 2 MODE
;P3M              .equ      %F7          ;PORT 3 MODE
;P01M             .equ      %F8          ;PORT 0-1 MODE

;IPR              .equ      F9H          ;INTERRUPT PRIORITY REGISTER
;IRQ              .equ      FAH          ;INTERRUPT REQUEST REGISTER
;IMR              .equ      %FB          ;INTERRUPT MASK REGISTER

;FLAGS            .equ      FCH          ;PROGRAM CONTROL FLAGS

;RP               .equ      FDH          ;PROGRAM CONTROL FLAGS

;SPL              .equ      %FF          ;STACK POINTER LSB

;SPH              .equ      %FE          ;STACK POINTER MSB  - CCP EMU ONLY

                      
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
;**                                                                         **
;**                    APPLICATION SPECIFIC VARIABLES                       **
;**        (Thresholds, time-outs, etc. which can be fine-tuned             **
;**                                                                         **
;*****************************************************************************
;*****************************************************************************
                                                                           
STOP_MINUS_SLOPE_TIME  .equ 150           ; Stop MINUS_SLOPE test x seconds  *
                                          ; after charging begins.           *
   
MINUS_DELTA_V_TRIP     .equ 5             ; A value of 7 represents a 0.5%   *
                                          ; decline in battery voltage.      *
   
DVDT_INF1_TRIP         .equ  15           ; First inflection threshold.      *
                                                                             
DVDT_INF2_TRIP         .equ  10           ; Second inflection threshold.     *
                                                                             
UNDERVOLTAGE_TEST_TIME .equ 5             ; Test duration, seconds.          *
                                                                             
AD_UNDERRANGE          .equ  722          ; Value 722 represents about 2 VDC *
                                                                             
MAX_HIZ_TEST_TIME      .equ  10           ; Test duration, seconds           *
                                                                             
TIME_LIMIT             .equ  2100         ; Time-out at X seconds total time *
                                          ; w/ charge current ON. 2100 = 35m *
                                                                             
DWELL_TIME             .equ  30           ; Dwell mode period, seconds       *
                                                                             
FFCD_TIME              .equ  34           ; Seconds after battery insertion  *
                                          ;  start NiCd fast full chg test   *
                                                                                                                                                                                     
EQUALIZE_TIME          .equ  120          ; Number of 1 second ON-intervals  *
                                          ; for post-charging Equalization   *
                                          ; mode. (for duty below, 2 hours)  *
                                                                             
EQUALIZE_DUTY          .equ  59           ; Equalization mode duty cycle:    *
                                          ; Current ON 1 sec, OFF xx sec     *
                                          ; For 1/60 x 1.6A pk = 27 mA avg   *
                                                                             
MAINT_DUTY             .equ  255          ; Maintenence mode duty cycle:     *
                                          ; Current ON 1 sec, OFF xxx sec    *
                                          ; For 1/256 x 1.6A pk = 6.2 mA avg *
                                                                             
PRESCALER_TARGET       .equ  3336         ; This value gives 175% headroom   *
                                          ; for battery A/D from start of    *
                                          ; charging NiCd to completion.     *
                                          ; Value corresponds to about 2.19V *
;*                                                                           *
;*****************************************************************************


;********************** CHARGING TERMINATION CODES ***************************
DVDT_INFLECTION_TERMINATION    .equ 2
MINUS_DELTA_V_TERMINATION      .equ 3   ;minus delta voltage test
TIMEOUT_TERMINATION            .equ 4
MINUS_SLOPE_TERMINATION        .equ 5
FFCD_TERMINATION               .equ 6   ;NiCd fast full charge detect

;***************************** PROBLEM CODES *********************************
UNDERVOLTAGE_PROBLEM           .equ 1
HIGH_IMPEDENCE_PROBLEM         .equ 2
OVERRANGE_W_I_ON_PROBLEM       .equ 3
INADVERTENT_TEST_MODE          .equ 4


;********************************* MACROS ************************************

LED_ON          .macro
                 AND      P2,#11110111B
                .endm


LED_OFF         .macro
                 OR       P2,#00001000B
                .endm


DELAY          .macro  A            ; Delay for "A" milliseconds.
                                    ; CYCLE TIME = (1/3579545 Hz) x 2 = 559 uS   
                                    ; DELAY LOOP = (27 CYCLES x 67)x559 uS x A

                LD     TEMP_MSB,#^HB (A*67) ;10 CYCLES
                LD     TEMP_LSB,#^LB (A*67) ;10 CYCLES

               .byte %5f ;wdt              ; 5 CYCLES * A
                DECW   TEMP_MSB             ;10 CYCLES * A
                JR     NZ,$-3               ;12 CYCLES * A
               .endm


XMIT_CHAR      .macro A 
                LD    SERIAL_OUTPUT_BYTE,A
                CALL  XMIT_BYTE
               .endm


XMIT_NUMERIC   .macro A
                LD    TEMP_MSB,A
                LD    TEMP_LSB,A+1
                CALL  SEND_NUMERIC_VALUE
               .endm


ADD_MINUS_DELTA_V_FACTOR: .macro  A
                LD  TEMP_MSB,A
                LD  TEMP_LSB,A+1
                ADD TEMP_LSB,#^LB MINUS_DELTA_V_TRIP
                ADC TEMP_MSB,#^HB MINUS_DELTA_V_TRIP
               .endm


LOAD:          .macro   A,B
                LD      A,B
                LD      A+1,B+1
               .endm


CLEAR:         .macro   A
                CLR     A
                CLR     A+1
               .endm


SUBTRACT:      .macro  A,B
                SUB    A+1,B+1
                SBC    A,B
               .endm


TEST:          .macro    A,OPERATOR,C,LABEL   ;operator = condition code
                CP       A,C                   ;compare first bytes
                JR       NE,$+5                ; not equal, so skip past rest
                CP       A+1,C+1               ; equal, so compare 2nd bytes
                JP       OPERATOR,LABEL        ;  jump if condition code true
               .endm


TEST2:         .macro    A,B,C,D
                CP       A,#^HB C
                JR       NE,$+5
                CP       A+1,#^LB C
                JP       B,D
               .endm


;******************* Program Address Space Begins Here ***********************
;**      First are the IRQ vectors, then the executeable code follows       **
;*****************************************************************************
                 .org  000H  
  
 .if CCP         
                 .word AD_CONVERT ;for CCP emulator, irq0 P32 rising/falling 
                                  ; edge. The IRQ reg will be loaded with bits 
                                  ; 7 & 6 as "01" for P32 rising edge trigger.
                                  ; - Z86C30 spec shows CCP interrupt scheme -
 .else           
                 .word POWER_UP1  ;otherwise irq0 P32 falling edge for Z86X07
 .endif

                 .word POWER_UP1  ;irq1 P33 falling edge
                 .word POWER_UP1  ;irq2 P31 falling edge
 
 .if CCP         
                 .word POWER_UP1  ;for CCP emulator, irq3 P30 falling edge int
                                  ; which is not available in the 07 devices!
 .else           
                 .word AD_CONVERT ;otherwise irq3 P32 rising edge for Z86X07
 .endif
                 
                 .word POWER_UP1  ;T0 interrupt vector (is polled this pgm)
                 .word POWER_UP1  ;irq5


;***************************** PROGRAM RESET *********************************
;**              Program execution starts here after reset                  **
;**     Execution begins here with a power_up or microcontroller WDT reset  **
;*****************************************************************************

POWER_UP:       NOP                    ;set breakpoint here to trap WDT resets
POWER_UP1:      DELAY  500             ;unused IRQs vector here. Delay 500 mS.

                TM     P2,#00001000B   ;CHECK IF TEST_MODE.
                JP     Z,TEST_MODE     ;TAKE JUMP IF IN TEST_MODE

                DELAY  750             ;DELAY AN ADDITIONAL 750 mS.

                LD    SPL,#80H         ;SETUP STACK POINTER
                CALL  INIT_SYSTEM_REGS ;INITALIZE REGISTERS & FLAGS
 
                LED_ON
                CALL  PIEZO_3          ;PIEZO BEEP FOR 3/32 SECOND
                CALL  DELAY_3          ;DELAY FOR 3/32 SECOND
                CALL  PIEZO_3          ;PIEZO BEEP AGAIN 
                LED_OFF


                      
;*******************************  SOFT START *********************************
;**     SOFT_START detects new battery insertions and initializes the       **
;**     system for a new charging cycle. The piezo sounds and the red led   **
;**     flashes as indications to the user of detected battery insertion.   **
;**     Detection of battery removal will branch back to this routine.      **
;*****************************************************************************

SOFT_START:      LD    SPL,#80H         ;SETUP STACK POINTER.
                 CALL  INIT_SYSTEM_REGS ;INITALIZE AND LOAD SYSTEM REGISTERS

                 CLEAR   TIMEOUT        ;ZERO OUT ACCUMULATED CHARGING TIME 
                 CLR     USER_FLAGS

                 LD    MODE,#'X'

                 CALL  GET_AD
                 CALL  SEND_SYSTEM_PARAMETERS
                 CALL  CHECK_FOR_BATTERY_REMOVAL

                 LD    MODE,#'B'        ;DEBOUNCE BATTERY DETECTION
                 CALL  GET_AD
                 CALL  SEND_SYSTEM_PARAMETERS
                 CALL  CHECK_FOR_BATTERY_REMOVAL

                 LED_ON
                 CALL  PIEZO_3

                 LD    MODE,#'I'
                 CALL  ONE_SECOND       ;  PRIOR TO DWELL CURRENT TURN ON



                      
;**************************** UNDERVOLTAGE TEST ******************************
;**   UNDERVOLTAGE_TEST is a routine which checks the relative battery      **
;**   A/D for a minimum value (which represents battery voltage over the    **
;**   constant, AD_UNDERRANGE. The routine executes in a loop with charge   **
;**   current ON for a max of UNDERVOLTAGE_TEST_TIME seconds.  Periodic     **
;**   battery A/D readings are taken with charge current OFF.  If battery   **
;**   voltage does not come into range within the test time, charging is    **
;**   prevented, & execution branches to Problem mode.                      **
;*****************************************************************************

UNDERVOLTAGE_TEST:
                 LD    MODE,#'U'

                 CALL  ONE_SEC_AD_SAMPLE
                 TEST2 AD, UGE, AD_UNDERRANGE, HIGH_Z_TEST

                 TEST2 TIME, ULT, UNDERVOLTAGE_TEST_TIME, UNDERVOLTAGE_TEST

                 LD    PROBLEM_MODE,#UNDERVOLTAGE_PROBLEM
                 JP    PROBLEM

                

                      
;*************************** HIGH IMPEDANCE TEST *****************************
;**                                                                         **
;**  HIGH_Z_TEST the resistive impedance characteristics of the battery     **
;**  A/D readings are taken with charge current on, then immediately after  **
;**  turning charge current off.  Value is derived from the current-OFF A/D **
;**  reading ( F(WO) ). If the current-ON A/D reading is less than F(WO),   **
;**  then HIGH_Z_TEST passes and program continues on to DWELL.             **
;**  Otherwise, this test is repeated in a loop until the battery tests OK  **
;**  or until MAX_HIZ_TEST_TIME runs out. If time-out occurs, a set of 8    **
;**  "beeps" is sounded on the piezo speaker at 1 second rate.  Also, the   **
;**  HI_Z_DETECTED flag is set. This forces a bypass of the                 **
;**  FULL_CHARGE_DETECTION routine, since a high impedance battery may not  **
;**  reliably respond to this test.  Program flow continues to DWELL mode.  **
;**  Circuit Assumptions:                                                   **
;**     .050 ohms is allowed for series contact resistance.                 **
;**     .020 ohms is allowed for series current sense resistance.           **
;**     .030 ohms at 1.2 volts per cell (current off) for battery voltage   **
;**       up to a maximum of 12 cells, NiCd.                                **
;**                                                                         **
;**   Using a curve fitting method, an approximation was derived.  This     **    
;**   approximation is the following linear equation.  To pass HI_Z_TEST:   **
;**                                                                         **
;**         A/D with current < (.9389 * A/D w/o current) + 465              **
;**         (.9389 is approximated by a more manageable 15/16)              **
;**                                                                         **
;*****************************************************************************

HIGH_Z_TEST:     LD    MODE,#'Z'
                 CALL  LED_CYCLE

                 LOAD  Z_TEST_I_ON,AD

                 CALL  DELAY_5
                 CALL  CURRENT_OFF
                 CALL  GET_AD                    ;1/32
                 CALL  CURRENT_ON
                 CALL  CHECK_FOR_BATTERY_REMOVAL

                 AND   USER_FLAGS,#01111111B   ;AD WAS TAKEN WITH CURRENT OFF,
                                               ;  CURRENT IS NOW ON. NEED TO 
                                               ; "TRICK" MODE SO THAT SERIAL
                                               ;  OUTPUT THINKS "z".
                 CALL  SEND_SYSTEM_PARAMETERS
                 OR    USER_FLAGS,#10000000B   ; "UNTRICK"

                 LOAD  TEMP_2,AD               ;TEMP = ((15 x AD_WO_I)/16)+465

                 LD    LOOP_COUNTER,#14
TIMES_15:
                 ADD   TEMP_LSB,AD_LSB
                 ADC   TEMP_MSB,AD_MSB
                 DEC   LOOP_COUNTER
                 JR    NE,TIMES_15

                 LD    LOOP_COUNTER,#4
DIVIDE_BY_16:
                 RCF
                 RRC   TEMP_MSB
                 RRC   TEMP_LSB
                 DEC   LOOP_COUNTER
                 JR    NE,DIVIDE_BY_16

                 ADD  TEMP_LSB,#^LB 465
                 ADC  TEMP_MSB,#^HB 465

                 TEST Z_TEST_I_ON, ULT, TEMP_2, DWELL

                 TEST2 TIME, ULT, MAX_HIZ_TEST_TIME, HIGH_Z_TEST

                 LED_ON
                 CALL   HI_Z_ALERT
                 LED_OFF
                 CALL   HI_Z_ALERT
                 OR     USER_FLAGS,#HI_Z_DETECTED
                 JR     DWELL


;     HI_Z_ALERT is a routine which is only used by the HI_Z_TEST routine.
;     HI_Z_ALERT beeps the piezo 4 times. Piezo is on for 2/32 second and
;     off for 2/32 of a second a total of 4 times. Time spent in this loop
;     is 1/2 second. 

HI_Z_ALERT:      LD     LOOP_COUNTER,#4
HI_Z_ALERT_LOOP: CALL   PIEZO_2
                 CALL   DELAY_2
                 DEC    LOOP_COUNTER
                 JR     NZ,HI_Z_ALERT_LOOP
                 RET



                      
;****************************** DWELL ROUTINE ********************************
;**  DWELL applies charge current for a period equal to DWELL_TIME seconds. **
;**  No charge termination tests are performed here, but battery removal is **
;**  checked.  This "do nothing" routine prevents false charge termination  **
;**  since battery voltage is often unpredictable when charge current is    **
;**  initially applied to a discharged battery.                             **
;*****************************************************************************

DWELL:
                 LD    MODE,#'D'

                 CALL  CURRENT_ON
                 CALL  ONE_SECOND    ;PERFORM "ONE_SECOND" ROUTINE

                 TEST2 TIME, ULT, DWELL_TIME, DWELL




                      
;*************************** AUTOSCALE ROUTINE *******************************
;**  AUTOSCALE changes the prescaler resistor divider circuit to provide    **
;**  the optimum attenuation value. There are 8 possible attenuation        **
;**  values. They are a binary progression of "00" through "07" with        **
;**  "00" being the greatest attenuation and "07" being no attenuation.     **
;**  The optimum attenuation is the pattern that provides the minimum       **
;**  attenuation while allowing suitable headroom for the battery           **
;**  voltage to climb during the charging cycle. The attenuator starts      **
;**  at the default maximum. The attenuation is iterativly decreased        **
;**  until the associated AD reading exceeds the PRESCALER_TARGET.          **
;**  The optimum attenuation value is the prescaler value just prior to     **
;**  exceeding PRESCALER_TARGET. We'll get as close to the target value     **
;**  as possible without exceeding it. Once the optimum attenuation is      **
;**  determined, the program flow continues to FULL_CHARGE_DETECTION.       **
;**  All AD  readings in AUTOSCALE are taken with charging current on.      **
;**                                                                         **
;**  A note about IGNORE_REMOVAL. Adjustment of the prescaler value may     **
;**  result in a false indication that a battery has been removed. For      **
;**  the time AUTOSCALE is active, we want to ignore removed battery        **
;**  indications. If a battery is removed during AUTOSCALE it will look     **
;**  like PRESCALER_TARGET is exceeded.  AUTOSCALE will think that it       **
;**  has found the optimum prescaler value and set it with the              **
;**  GOT_ATTEN code. The suspected optimum attenuation is set, the          **
;**  IGNORE_REMOVAL flag is cleared, and flow continues for ONE_SECOND.     **
;**  Here, the determination is made as to the battery removal during       **
;**  AUTOSCALE. Thus, if a battery is removed during AUTOSCALE, it may      **
;**  take up to 1.25 seconds to detect it instead of the typical            **
;**  maximum .25 seconds during any other portion of the charge cycle.      **
;*****************************************************************************

AUTOSCALE:        LD    MODE,#'A'
                  LD    ATTENUATOR,#01H
                  OR    USER_FLAGS,#IGNORE_REMOVAL

AUTOSCALE_LOOP:   CALL  PUT_ATTENUATOR_AND_SAMPLE
                  TEST2 AD, UGT, PRESCALER_TARGET, GOT_ATTEN ;also test that
                  INC   ATTENUATOR                           ;batt is present 
                  CP    ATTENUATOR,#8
                  JR    NE,AUTOSCALE_LOOP

GOT_ATTEN:        DEC   ATTENUATOR
                  AND   USER_FLAGS,#11011111B ;STOP IGNORE_REMOVAL.
                  CALL  PUT_ATTENUATOR_AND_SAMPLE
                  JR    FULL_CHARGE_DETECTION



PUT_ATTENUATOR_AND_SAMPLE:
                 CALL  PUT_ATTENUATOR
                 CALL  ONE_SECOND
                 RET


                      
;*********************** FAST FULL CHARGE DETECTION **************************
;**  This technique is used with NiCd batteries. This was developed to take **
;**  advantage of an observed relationship between state-of-charge and the  **
;**  change in battery voltage when charging current turns off.  See the    **
;**  full description in the software explanation.  The applicability to    **
;**  all battery types is not assured.  The leading semicolons (;) below    **
;**  indicates that this routine is ignored by the assembler.  Remove the   **
;**  semicolons to include the Fast Full Charge Detection routine.          ** 
;*****************************************************************************

FULL_CHARGE_DETECTION:
                 TM    USER_FLAGS,#HI_Z_DETECTED
                 JR    NE,CHARGING

;                 LD    TEMP,P0
;                 AND   TEMP,#03H

;                 CP   TEMP,#00H
;                 JR   EQ,CHARGING

;                 LD   FFCD_THRESHOLD_LSB,#120
;                 CP   TEMP,#01H
;                 JR   EQ,FFCD_ACTIVE
;
;                 LD   FFCD_THRESHOLD_LSB,#135
;                 CP   TEMP,#02H
;                 JR   EQ,FFCD_ACTIVE

;                 LD   FFCD_THRESHOLD_LSB,#150

;FFCD_ACTIVE:      LD    MODE,#'S'

;                 CALL  ONE_SECOND    ;PERFORM "ONE_SECOND" ROUTINE

;                 CP    TIME_LSB,#FFCD_TIME
;                 JR    LT,FFCD_ACTIVE

;                 CALL  LED_CYCLE
;                 CALL  LED_OFF_CYCLE

;                 CALL  CURRENT_OFF    ;current off and immediatly take
;                 CALL  GET_AD         ;  an A/D reading.

;                 LOAD    FFCD_FACTOR,AD    ;use A/D reading as 1ST FFCD_FACTOR.

;                 CALL  SEND_SYSTEM_PARAMETERS ;2/32 communicate it.

;                 LD    LOOP_COUNTER,#5  ;perform ONE_SECOND routine 10
;FFCD_LOOP_I_OFF:
;                 CALL  ONE_SECOND       ;  times.
;                 DEC   LOOP_COUNTER
;                 JR    NE,FFCD_LOOP_I_OFF

;                 SUBTRACT  FFCD_FACTOR,AD  ;FFCD_FACTOR holds the
;                 LOAD      AD,FFCD_FACTOR
;                 CALL      SEND_SYSTEM_PARAMETERS ;2/32 communicate it.
;                 TEST      FFCD_FACTOR, ULT, FFCD_THRESHOLD, CHARGING
;                 LD        INFLECTION,#FFCD_TERMINATION
;                 JP        TERMINATE


;********************** MAIN BATTERY CHARGING ROUTINE ************************
;**  Z8 execution will loop through this routine while battery charging is  **
;**  active. When full charge is reached, program branches to TERMINATE.    **
;*****************************************************************************

CHARGING:
    LD    MODE,#'Q'                              ;Signal outside serial data
                                                 ; rcvr that all termination 
                                                 ; tests are performed after 
                                                 ; the next battery a/d.
    CALL  ONE_SEC_AD_SAMPLE                      ;Sample battery voltage (with
                                                 ; current off) & send serial
                                                 ; data.  Return in 1 second.
    CALL  MINUS_DELTA_V_TEST                     ;Minus delta voltage test.
    CALL  STACK_AD                               ;Push new a/d onto stack.
    CALL  DVDT_INFL_TEST                         ;Inflection analysis tests.
    CALL  MINUS_SLOPE_TEST                       ;Minus slope test.
    LD    LOOP_COUNTER,#3                        ;Initialize loop counter.
CHARGING_LOOP:
    CALL  CHARGING_3_SECONDS                     ;3 second delay w/current on.
    LD    MODE,#'Y'                              ;Only minus delta volt test.
                                                 ; for three consecutive a/d's
    CALL  ONE_SEC_AD_SAMPLE                      ;Sample battery voltage (with
                                                 ; current off) & send serial
                                                 ; data.  Return in 1 second.
    CALL  MINUS_DELTA_V_TEST                     ;Minus delta voltage test.
    DEC   LOOP_COUNTER                           ;Loop 3 times, then... 
    JR    NE,CHARGING_LOOP

    CALL  CHARGING_3_SECONDS                     ;3 second delay w/current on.
    JR    CHARGING                               ;Loop to top for all tests 
                                                 ; after the next a/d sample.

CHARGING_3_SECONDS:
                 LD    MODE,#'C'
                 CALL  ONE_SECOND
                 CALL  ONE_SECOND
                 CALL  ONE_SECOND
                 RET


                      
;************************** TERMINATION ROUTINE ******************************
;**         Execution branches here when battery charging is finished       **
;*****************************************************************************

TERMINATE:
    LD     MODE,#'F'         ;mode is "FINISHED".
    CALL   CURRENT_OFF       ;turn CURRENT_OFF.
    LED_ON                   ;turn LED_ON solid.

    CALL   SEND_SYSTEM_PARAMETERS  ;communicate charging termination.

    CALL   SET_MAXIMUM_ATTENUATOR
                 
    CALL   PIEZO_5           ;BEEP...
    CALL   DELAY_5 
    CALL   PIEZO_5           ;       BEEP...
    CALL   DELAY_5
    CALL   PIEZO_5           ;              BEEP.     



                      
;*************************** EQUALIZATION MODE *******************************
;**  Trickle charge briefly to "top off" any cells in battery pack that are **
;**    not fully charged.  Since termination tests act on overall battery   **
;**     response, it's possible that some cells will still accept charge.   **
;*****************************************************************************

EQUALIZATION:
     LD     MODE,#'E'                    ;load mode register for EQUALIZATION

     CLEAR  TIME

EQUALIZE_LOOP:
     LD     LOOP_COUNTER,#EQUALIZE_DUTY  ; establish EQUALIZATION mode's   
                                         ;  charge current duty cycle.
     CALL   EQUAL_MAINT_DUTY_CYCLE       ; perform one EQUALIZATION current
                                         ;  duty cycle.
     TEST2  TIME, ULT, EQUALIZE_TIME, EQUALIZE_LOOP 
                                         ;  after programmed eq mode duration
                                         ;  fall through to MAINTENANCE mode.

                      
;***************************** MAINTENANCE MODE ******************************
;**    Now revert to very low charge rate to compensate for self discharge. **
;**  Select duty cycle which will not cause problem for indefinite period.  **
;*****************************************************************************

MAINTENANCE:     
     LD     MODE,#'M'                    ;load mode register for MAINTENANCE
     CLEAR  TIME

MAINTENANCE_LOOP:
     LD     LOOP_COUNTER,#MAINT_DUTY     ;establish MAINTENANCE mode's 
                                         ; charge current duty cycle.
     CALL   EQUAL_MAINT_DUTY_CYCLE       ;perform one MAINTENANCE current
                                         ; duty cycle.
     JR     MAINTENANCE_LOOP             ;loop forever in MAINTENANCE mode,
                                         ; or until battery removal detected.


                      
;***************************** PROBLEM MODE **********************************
;**      Execution branches here if charger detects problem condition.      **
;*****************************************************************************

PROBLEM          CALL   CURRENT_OFF
                 LED_OFF
                 AND    USER_FLAGS,#%FF-IGNORE_REMOVAL
                 CALL   SEND_SYSTEM_PARAMETERS
                 CALL   SET_MAXIMUM_ATTENUATOR

                 LED_ON
                 LD     MODE,#'P'
                 LD     INFLECTION,PROBLEM_MODE
                 CALL   SEND_SYSTEM_PARAMETERS

PROBLEM_LOOP     LED_OFF
                 CALL   DELAY_1        ;1/32
                 CALL   GET_AD         ;1/32
                 CALL   CHECK_FOR_BATTERY_REMOVAL

                 LED_ON
                 CALL   PIEZO_2
                 JR     PROBLEM_LOOP

;****************************** SUBROUTINES **********************************
;*****************************************************************************
                      
ONE_SECOND:      

ONE_SECOND_CYCLE: CALL  LED_CYCLE
                  CALL  LED_OFF_CYCLE
                  RET


ONE_SEC_AD_SAMPLE:
                 CALL  CURRENT_ON

                 CALL  LED_CYCLE
                 CALL  DELAY_4                   ;4/32
                 CALL  CURRENT_OFF
                 CALL  DELAY_1                   ;1/32
                 CALL  GET_AD                    ;1/32
                 CALL  CURRENT_ON
                 CALL  CHECK_FOR_BATTERY_REMOVAL

                 AND   USER_FLAGS,#01111111B ;AD WAS TAKEN WITH CURRENT OFF,
                                             ;  CURRENT IS NOW ON. NEED TO 
                                             ; "TRICK" MODE SO THAT SERIAL
                                             ;  OUTPUT THINKS "c".
                 CALL  SEND_SYSTEM_PARAMETERS     ;2/32
                 OR    USER_FLAGS,#10000000B  ;"UNTRICK"
                 RET


                      
LED_CYCLE:       LED_ON
                 CALL  INC_TIME
                 CALL  GET_AD_CYCLE
                 CALL  GET_AD_CYCLE
                 
LED_OFF_CYCLE:   LED_OFF
                 CALL  GET_AD_CYCLE
                 
                 RET


INC_TIME:        ADD   TIME_LSB,#01H
                 ADC   TIME_MSB,#00H

                 TM    USER_FLAGS,#CURRENT_STATE
                 JR    EQ,TIME_OK

                 ADD   TIMEOUT_LSB,#01H
                 ADC   TIMEOUT_MSB,#00H

                 TEST2 TIMEOUT, ULT, TIME_LIMIT, TIME_OK  

                 LD    INFLECTION,#TIMEOUT_TERMINATION
                 JP    TERMINATE

TIME_OK:         RET

                      
EQUAL_MAINT_DUTY_CYCLE:
                 CALL   CURRENT_OFF
EQUAL_MAINT_DUTY_OFF:
                 CALL   EQUAL_MAINT_1_SECOND
                 DEC    LOOP_COUNTER
                 JR     NE,EQUAL_MAINT_DUTY_OFF

                 CALL   CURRENT_ON

EQUAL_MAINT_1_SECOND:
                 CLEAR  TIMEOUT
                 call   INC_TIME
                 CALL   EQUAL_MAINT_HALF_SECOND

EQUAL_MAINT_HALF_SECOND:
                 CALL   GET_AD_CYCLE
                 CALL   GET_AD_CYCLE
                 RET

                     
DELAY_5          CALL  DELAY_1
DELAY_4          CALL  DELAY_1
DELAY_3          CALL  DELAY_1
DELAY_2          CALL  DELAY_1
DELAY_1          
                .byte %5f ;wdt
                 TM    IRQ,#10H     ;IS T0 TIMED OUT?
                 JR    Z,DELAY_1
                 AND   IRQ,#%EF     
                 RET

                 
                      
GET_AD_CYCLE     .byte %5f                       ;WDT (reset watchdog timer)
                 CALL  DELAY_5                   ;5/32
                 CALL  GET_AD                    ;1/32
                 CALL  SEND_SYSTEM_PARAMETERS    ;2/32
                 CALL  CHECK_FOR_BATTERY_REMOVAL
                 RET

;**********************  LOG A/D CONVERSION SUBROUTINE ***********************
;**  Return with measured battery log a/d value in AD_MSB / AD_LSB register **
;*****************************************************************************

GET_AD           .byte %5f                ;NEW WDT
                 AND   P2,#%EF            ;BEGIN A/D CONVERSION
                 SRP   #10H               ; 6

                 LD    TEMP,#AD_MSB_0     ;10
INIT_AD_MSB_LOOP:
                 LD    @TEMP,#%FF         ;10*16  CLEAR LOCATION POINTER TO BY 7DH.
                 INC   TEMP               ; 6*16  POINT TO NEXT RAM LOCATION.
                 CP    TEMP,#(AD_MSB_F+1) ;10*16  CHECK IF CLEAR LOOP HITS TOP.
                 JR    NE,INIT_AD_MSB_LOOP;12*16  LOOP IF NOT.

                 LD    AD_MSB,#10H        ;10
                 CLR   AD_LSB             ; 6

                 LD    TEMP,#159          ;10

DELAY_FOR_COMMON_MODE:
                .byte %5f                 ;wdt  5*159

                 DEC   TEMP                         ; 6*159
                 JR    NE,DELAY_FOR_COMMON_MODE     ;12*159
                                                    ;CYCLES = 4314 = 2.41 MS

                 ;OR   P2,#01000000B  ;Diagnostic tick: gen signal that can be
                                   ;     seen w/ scope during A/D conversion.  
                                   ;Sets piezo pin p26. resets at GOT_AD below
                                   ;Trigger scope on falling edge of p24. Then
                                   ;   p26 r-edge = end of common mode delay
                                   ;   p26 f-edge = A/D comparator IRQ 
                                   ;Tick sound reminds you: take this out!

                 .byte %5f                          ;wdt  5*159 
                 TM    P3,#04H                      ;Is VBAT > log RC ramp? 10
                 JR    NE,SYNC_AD_RET               ; yes, error condition  10

                 CLR   AD_MSB                       ;6

 .if CCP         
                 AND   IRQ,#%3E   ;RESET IRQ0 INTERRUPT REQUEST           10
                 OR    IRQ,#%40   ;SET EDGE LIKE E07/C07 (P31=FALL, P32=RISE)
                 LD    IMR,#%81   ;ENABLE P32 RISING EDGE INTERRUPT IRQ0  10 

 .else           
                 AND   IRQ,#%37   ;RESET IRQ3 INTERRUPT REQUEST           10 
                 LD    IMR,#%88   ;ENABLE P32 RISING EDGE INTERRUPT IRQ3  10 
 
 .endif



                 EI                       ;6
                 DJNZ  R15,$   ;NOT CORRECT HERE ON DOWN. 
                 DJNZ  R14,$   ;(255*12)+10  7740-10810      6.039 ms
                 DJNZ  R13,$   ;(255*12)+10  10810-13880+6   7.636 ms
                 .byte %5f ;wdt
                 DJNZ  R12,$   ;(255*12)+10  13886-16956     9.473 ms
                 DJNZ  R11,$   ;(255*12)+10  16956-20026    11.188 ms
                 DJNZ  R10,$   ;(255*12)+10  20026-23096+6  12.906 ms
                 .byte %5f ;wdt           ;5
                 DJNZ  R9,$    ;(255*12)+10  23102-26172    14.621 ms
                 DJNZ  R8,$    ;(255*12)+10  26172-29242    16.336 ms
                 DJNZ  R7,$    ;(255*12)+10  29242-32312+6  18.054 ms
                 .byte %5f ;wdt           ;6
                 DJNZ  R6,$    ;(255*12)+10  32318-35388    19.770 ms
                 DJNZ  R5,$    ;(255*12)+10  35388-38458    21.485 ms
                 DJNZ  R4,$    ;(255*12)+10  38458-41528+6  23.203 ms
                 .byte %5f ;wdt           ;6
                 DJNZ  R3,$    ;(255*12)+10  41534-44604    24.918 ms
                 DJNZ  R2,$    ;(255*12)+10  44604-47674    26.634 ms
                 DJNZ  R1,$    ;(255*12)+10  47674-50744+6  28.352 ms
                 .byte %5f ;wdt           ;6
                 DJNZ  R0,$    ;(255*12)+10  50750-53820    30.067 ms
      
SYNC_AD_RET      DI

                 CALL  DELAY_1

                 OR    P2,#10H

                 RET


AD_CONVERT       .byte %5f             ;WDT added

                 LD    AD_MSB,#1EH     ;USE AD_MSB AS POINTER TO LSB STACK.
                 
FIND_AD_MSB      CP    @AD_MSB,#%FF    ;CHECK CONTENTS OF POINTER = FF?
                 JR    EQ,GOT_AD
                 DEC   AD_MSB
                 CP    AD_MSB,#0FH
                 JR    NE,FIND_AD_MSB

GOT_AD           INC   AD_MSB
                 ;AND  P2,#10111111B ;Diagnostic tick: ends here, reset p26.
                                     ;This terminates oscilloscope output
                                     ;begun earlier in GET_AD. See note above.
                 LD    AD_LSB,@AD_MSB
                 AND   AD_MSB,#0FH

                 LD    TEMP,#0FH
                 SUB   TEMP,AD_MSB
                 ADD   AD_LSB,TEMP
                 ADC   AD_MSB,#00H

                 
                 

                 IRET

               
;*************************** STACK_AD SUBROUTINE  ****************************
;**    Push new A/D reading onto 8 level FIFO A/D stack at register AD0.    **
;**                   Lose oldest A/D in register AD7.                      **
;**       In 4C charger, this subroutine is called every 16 seconds.        **
;*****************************************************************************

STACK_AD:
          LOAD  AD7,AD6
          LOAD  AD6,AD5
          LOAD  AD5,AD4
          LOAD  AD4,AD3
          LOAD  AD3,AD2
          LOAD  AD2,AD1
          LOAD  AD1,AD0
          LOAD  AD0,AD
          RET
                
;******************* DVDT INFLECTION TEST SUBROUTINE *************************
;**          Here is where the inflection analysis takes place              **
;*****************************************************************************

DVDT_INFL_TEST:
          TEST2 AD7, NE, 0, DVDT_INFL_TEST_ACTIVE
          RET

DVDT_INFL_TEST_ACTIVE:
          CP   INFLECTION,#1
          JR   EQ,CHECK_FOR_SECOND_INFLECTION


;slope=AD0MSB,AD0LSB+AD1MSB,AD1LSB+AD2MSB+AD2LSB+AD3MSB+AD3LSB-
;     (AD4MSB,AD4LSB+AD5MSB,AD5LSB+AD6MSB,AD6LSB+AD7MSB,AD7LSB)
          
          LOAD  SLOPE,AD0

          ADD  SLOPE_LSB,AD1LSB
          ADC  SLOPE_MSB,AD1MSB

          ADD  SLOPE_LSB,AD2LSB
          ADC  SLOPE_MSB,AD2MSB

          ADD  SLOPE_LSB,AD3LSB
          ADC  SLOPE_MSB,AD3MSB

          SUBTRACT  SLOPE,AD4

          SUBTRACT  SLOPE,AD5

          SUBTRACT  SLOPE,AD6

          SUBTRACT  SLOPE,AD7

;IF (SLOPE_MSB >= 0) then jump DVDT_INFL_TEST else ZERO SLOPE
          JR   PL,FIRST_INFLECTION_SLOPE_OK
          CLEAR  SLOPE

FIRST_INFLECTION_SLOPE_OK:
          TEST  SLOPE, UGT, MSLOPE, MINIMUM_SLOPE_OK

          LOAD  MSLOPE,SLOPE

          RET

MINIMUM_SLOPE_OK:
;temp = MSLOPE_MSB,MSLOPE_LSB + DVDT_INF1_TRIP_MSB,DVDT_INF1_TRIP_LSB
          LOAD TEMP_2,MSLOPE
          ADD  TEMP_LSB,#^LB DVDT_INF1_TRIP
          ADC  TEMP_MSB,#^HB DVDT_INF1_TRIP

;
;IF slope > (mslope+dvdt_trip) then GOT_1ST_INFLECTION
          TEST SLOPE, UGT, TEMP_2, GOT_1ST_INFLECTION
          RET

;else inflection = 1 and mslope=slope
GOT_1ST_INFLECTION:
          LD   INFLECTION,#1

          CLEAR  MSLOPE


CHECK_FOR_SECOND_INFLECTION:
;slope=( AD0 + AD1 ) - ( AD3 + AD4 )
          LOAD   SLOPE,AD0

          ADD  SLOPE_LSB,AD1LSB
          ADC  SLOPE_MSB,AD1MSB

          SUBTRACT SLOPE,AD3

          SUBTRACT SLOPE,AD4

;IF (SLOPE_MSB < 0) then SLOPE = 0 
          JR   PL,SECOND_INFLECTION_SLOPE_OK
          CLEAR  SLOPE

SECOND_INFLECTION_SLOPE_OK:
          TEST  SLOPE, ULT, MSLOPE, MAXIMUM_SLOPE_OK
          LOAD  MSLOPE,SLOPE
          RET

MAXIMUM_SLOPE_OK:
;temp = MSLOPE_MSB,MSLOPE_LSB - dvdt_inf2_trip
          LOAD  TEMP_2,MSLOPE

          SUB  TEMP_LSB,#^LB DVDT_INF2_TRIP
          SBC  TEMP_MSB,#^HB DVDT_INF2_TRIP

;IF slope < (mslope - dvdt_trip) then GOT_DVDT_INFL_TERMINATION
          TEST SLOPE, ULT, TEMP_2, GOT_DVDT_INFL_TERMINATION
          RET

GOT_DVDT_INFL_TERMINATION:
          LD   INFLECTION,#DVDT_INFLECTION_TERMINATION
          JP   TERMINATE

;************************** MINUS SLOPE TEST SUBROUTINE **********************
;**  This test is performed during two windows in a normal charge cycle:    **
;**  (a) after charging commences for a program-specified time interval,    **
;**  and (b) after first inflection has occurred.  The flat part of charge  **
;**  cycle is deactivated.  Register AD0 contains the newest A/D value.     **
;**  If (AD0 < AD1) and (AD1 < AD2) then dV/dt slope is negative, so        **
;**  charging termination will occur.                                       **
;*****************************************************************************

MINUS_SLOPE_TEST:
          TEST2 TIME, ULT, STOP_MINUS_SLOPE_TIME, MINUS_SLOPE_ACTIVE 
          CP    INFLECTION,#1                ;check for start time & inflection
          JR    EQ,MINUS_SLOPE_ACTIVE        ;  continue if test valid

          RET

MINUS_SLOPE_ACTIVE:
          TEST AD0, UGE, AD1, MINUS_SLOPE_RET ; test three most recent A/D values

          TEST AD1, UGE, AD2, MINUS_SLOPE_RET

MINUS_SLOPE:
          LD  INFLECTION,#MINUS_SLOPE_TERMINATION
          JP  TERMINATE

MINUS_SLOPE_RET:
          RET


;******************** Minus Delta Voltage Test Subroutine ********************
;**  Do not confuse with minus dV/dt slope test.  This checks if two most   **
;**  recent A/D readings are less than stored max A/D value by threshold    **
;**  amount.  Special 2-level stack is updated every 4 sec for  4C charger. **
;*****************************************************************************


MINUS_DELTA_V_TEST:

          LOAD   AD1_MINUS_DELTA_V,AD0_MINUS_DELTA_V

          LOAD   AD0_MINUS_DELTA_V,AD

; If (AD0 > MAXAD) and (AD1 > MAXAD) then register MAXAD is less then 
; both registers AD0 & AD1.  Check:

          TEST MAXAD, UGT, AD0_MINUS_DELTA_V, TEST_FOR_MINUS_DELTA_V
          TEST MAXAD, UGT, AD1_MINUS_DELTA_V, TEST_FOR_MINUS_DELTA_V

          TEST AD1_MINUS_DELTA_V, UGT, AD0_MINUS_DELTA_V, MAXAD_IS_AD0_MINUS_DELTA_V
          LOAD MAXAD,AD1_MINUS_DELTA_V
          RET

MAXAD_IS_AD0_MINUS_DELTA_V:
          LOAD   MAXAD,AD0_MINUS_DELTA_V
          RET

TEST_FOR_MINUS_DELTA_V:
; If (AD0_F+MINUS_DELTA_V_TRIP < MAXAD) and (AD1+MINUS_DELTA_V_TRIP < MAXAD) 
; then terminate for minus delta voltage.     First check AD0+trip < MAXAD ? 

          ADD_MINUS_DELTA_V_FACTOR AD0_MINUS_DELTA_V   
          TEST TEMP_2, ULT, MAXAD, CHECK_AD1_MINUS_DELTA_V ;yes, so check AD1
          RET                                              ;no, bail out

CHECK_AD1_MINUS_DELTA_V:
          ADD_MINUS_DELTA_V_FACTOR AD1_MINUS_DELTA_V       ;Now check AD1  
          TEST TEMP_2, ULT, MAXAD, GOT_MINUS_DELTA_V_TERMINATION 
          RET

GOT_MINUS_DELTA_V_TERMINATION:
          LD   INFLECTION,#MINUS_DELTA_V_TERMINATION
          JP   TERMINATE

;************************* Piezo Sound Subroutines ***************************
;**      Simple bit banging to generate square wave piezo beeps             **
;*****************************************************************************

PIEZO_5          CALL  PIEZO_1      ;entry here PIEZO on for 5/32 second.
PIEZO_4          CALL  PIEZO_1      ;entry here PIEZO on for 4/32 second.
PIEZO_3          CALL  PIEZO_1      ;entry here PIEZO on for 3/32 second.
PIEZO_2          CALL  PIEZO_1      ;entry here PIEZO on for 2/32 second.
PIEZO_1                             ;entry here PIEZO on for 1/32 second.

PIEZO_BIT_HIGH:  OR    P2,#01000000B   ;10  drive PIEZO bit port 2 high.
                 CALL  PIEZO_DELAY     ;20 + 173 timing delay.
                 .byte %5f ;wdt                   ;5
                 .byte %5f ;wdt                   ;5
                 NOP                   ;6
                 NOP                   ;6

PIEZO_BIT_LOW:   AND   P2,#10111111B   ;10  drive PIEZO bit port 2 low.
                 CALL  PIEZO_DELAY     ;20 + 173 timing delay.
                 TM    IRQ,#10H        ;10  has T0 timed out?
                 JR    Z,PIEZO_BIT_HIGH;12  no, continue loop until it does.
                 AND   IRQ,#%EF        ;    else, clear IRQ flag and
                 RET                   ;  RETURN

PIEZO_DELAY:
                 .byte %5f              ;wdt   ;5
                 PUSH   TEMP           ;10
                 LD     TEMP,#7        ;10
PIEZO_DELAY_LOOP:
                 DEC    TEMP           ;6 * 7
                 JR     NE,PIEZO_DELAY_LOOP ; (12 * 6) + 10
                 POP    TEMP           ;10
                 RET                   ;14 TOTAL 173


;******************** Serial Data Dispatch Subroutine ************************
;**       This sends charging parameters to external hardware.              **
;*****************************************************************************

SEND_SYSTEM_PARAMETERS:
                 AND       MODE,#11011111B  ;DEFAULT MODE UPPERCASE LETTER.
                 TM        USER_FLAGS,#10000000B
                 JR        NE,SEND_MODE_CHAR
                 OR        MODE,#00100000B  ;MODE IS LOWERCASE

SEND_MODE_CHAR
                 XMIT_CHAR MODE

                 XMIT_CHAR #' '
                 LD   TEMP,CONFIGURATION
                 ADD  TEMP,#'0'
                 XMIT_CHAR TEMP

                 XMIT_NUMERIC TIME

                 XMIT_NUMERIC TIMEOUT

                 XMIT_NUMERIC AD

                 XMIT_NUMERIC SLOPE

                 XMIT_CHAR #' '
                 LD   TEMP,INFLECTION
                 ADD  TEMP,#'0'
                 XMIT_CHAR TEMP

                 XMIT_NUMERIC MSLOPE

                 XMIT_NUMERIC MAXAD

                 XMIT_CHAR #' '
                 LD   TEMP,ATTENUATOR
                 ADD  TEMP,#'0'
                 XMIT_CHAR TEMP

                 CALL SEND_CR_LF

                 CALL DELAY_2

                 RET

SEND_NUMERIC_VALUE:
                 XMIT_CHAR #' '

                 LD    DIVISOR_MSB,#^HB 10000
                 LD    DIVISOR_LSB,#^LB 10000
                 CALL  SEND_NUMERIC_VALUE_DIGIT

                 LD    DIVISOR_MSB,#^HB 1000
                 LD    DIVISOR_LSB,#^LB 1000
                 CALL  SEND_NUMERIC_VALUE_DIGIT

                 CLR   DIVISOR_MSB
                 LD    DIVISOR_LSB,#^LB 100
                 CALL  SEND_NUMERIC_VALUE_DIGIT

                 CLR   DIVISOR_MSB
                 LD    DIVISOR_LSB,#^LB 10
                 CALL  SEND_NUMERIC_VALUE_DIGIT

                 LD    SERIAL_OUTPUT_BYTE,TEMP_LSB
                 ADD   SERIAL_OUTPUT_BYTE,#'0'
                 CALL  XMIT_BYTE

                 RET


SEND_NUMERIC_VALUE_DIGIT:
                 LD  SERIAL_OUTPUT_BYTE,#%FF

SNV_DIGIT_LOOP:  INC  SERIAL_OUTPUT_BYTE
                 SUBTRACT  TEMP_2,DIVISOR
                 JR   PL,SNV_DIGIT_LOOP

                 ADD  TEMP_LSB,DIVISOR_LSB
                 ADC  TEMP_MSB,DIVISOR_MSB

                 ADD  SERIAL_OUTPUT_BYTE,#'0'
                 CALL XMIT_BYTE

                 RET


;******  Transmit serial data on P02, pin 12.  (Z86E08 OTP & ICE box)  ******

XMIT_BYTE:
    ld   TEMP,#10            ;  1 start bit, 8 data bits, 1 stop bits.
    rcf                      ;  create start bit by setting C low.
    rlc  SERIAL_OUTPUT_BYTE  ;  7 lower signifigant bits are rotated left
                             ;  out of the way with the cleared carry
                             ;  bit going into D0 position. D0 will 
                             ;  serve as start bit. The D7 bit ends up 
                             ;  in the C flag and is recovered in the 
                             ;  very next instruction.

XMIT_LOOP:
    rrc  SERIAL_OUTPUT_BYTE  ;rotate right once getting D0 into C. 
                             ;  Prepare to xmit it.
    jr   C,XMIT_HIGH_BIT     ;Take jump if C high to send it as a high.

XMIT_LOW_BIT:
     cp   MODE,#'T'             ;  C is low, are we in TEST_MODE?
     jr   EQ,DIRECT_TO_LED_LOW  ;  yes, take jump to direct output low to LED pin
     ld   P01M,#04H             ;  no, didn't get here from TEST_MODE, P02 is 
                                ;  already low, this makes it an output.
     jr   BIT_DELAY             ;  take jump to delay routine. 

DIRECT_TO_LED_LOW:
     and  P2,#11110111B         ;in TEST_MODE, LED port low.
     jr   BIT_DELAY             ;take jump to delay routine.

XMIT_HIGH_BIT:
     cp   MODE,#'T'             ;C is high, are we in TEST_MODE?
     jr   EQ,DIRECT_TO_LED_HIGH ;yes, take jump to direct output high to LED
     ld   P01M,#05H             ;no, didn't get here from TEST_MODE, P02 is 
                                ; made input for high impedance.
     jr   BIT_DELAY             ;take jump to delay routine. 

DIRECT_TO_LED_HIGH:
      or   P2,#00001000B    ;in TEST_MODE, led port high.
      jr   BIT_DELAY        ;take jump to delay routine.

BIT_DELAY:
      push TEMP             ;calling routine may be using TEMP.
      ld   TEMP,#4          ;TEMP will be a loop counter.
BIT_DELAY_LOOP:
      dec  TEMP             ;dec TEMP
      jr   NE,BIT_DELAY_LOOP;  and loop till TEMP exhausted.
      pop  TEMP             ;recover TEMP.

      scf                   ;this so that SERIAL_OUTPUT_BYTE becomes
                            ;  flushed with 1's which serve conviently
                            ;  as stop bits.
      dec  TEMP             ;keeps track of number of rotations.
      jr   NE,XMIT_LOOP     ;take jump till all bits are sent.

      .byte %5f ;wdt
      ret



                      
SEND_CR_LF       XMIT_CHAR   #%0D
                 XMIT_CHAR   #%0A
                 RET

                      
;******************* CHARGE CURRENT HANDLER SUBROUTINES **********************
;**    These turn charging current on/off and set/reset the flag.           **
;*****************************************************************************

CURRENT_ON:      OR       P2,#00100000B         ;CHARGE ENABLE BIT HIGH
                 OR       USER_FLAGS,#10000000B ;SET FLAG
                 RET

CURRENT_OFF:     AND      P2,#11011111B         ;CHARGE ENABLE BIT LOW
                 AND      USER_FLAGS,#01111111B ;RESET FLAG
                 RET

;*************** REGISTER & FLAG INITIALIZATION SUBROUTINE *******************
;**    This is executed at reset, battery removal, or during in test mode   **
;*****************************************************************************
                      
INIT_SYSTEM_REGS:

 .if CCP         
                 LD    SPH,#00H         ;CCP EMULATOR ONLY, ZERO STK PTR HI 
 .else

 .endif
                 LD    7DH,#07H         ;LOCATION 7DH USED AS POINTER. 
                                        ;  INITALIZED AS 07.
CLR_RAM_LOOP:    CLR   @7DH             ;CLEAR LOCATION POINTED TO BY 7DH.
                 INC   7DH              ;POINT TO NEXT RAM LOCATION.
                 CP    7DH,#7DH         ;CHECK IF CLEAR LOOP HITS TOP.
                 JR    NE,CLR_RAM_LOOP  ;LOOP IF NOT.
                 CLR   7DH              ;CLEAR POINTER ITSELF.

                 LD    P01M,#05H   ;PORT 0 AND 1 MODE SELECT
                                   ;  INTERNAL STACK SELECTED
                                   ;  PORT 0 IS INPUT

                 LD   CONFIGURATION,P0
                 LD   P0,CONFIGURATION
                 AND  P0,#%FB      ;P0 BIT 2 LOW FOR SERIAL OUTPUT

                 AND  CONFIGURATION,#07H

                                   
                 LD    P3M,#03H    ;PORT MODE REGISTER 0000 0011
                                   ;  PORT 3 ANALOG MODE
                                   ;  PORT 2 ACTIVE PULL-UPS

                 LD    P2,#%98    ;  BIT 7 ZERO-X INPUT           HIGH
                                  ;  BIT 6 PIEZO OUTPUT           LOW
                                  ;  BIT 5 CHARGE ENABLE OUTPUT   LOW
                                  ;  BIT 4 AD ARM/CONVERT OUTPUT  HIGH

                                  ;  BIT 3 LED OUTPUT             HIGH
                                  ;  BIT 2 PRESCALER OUTPUT       LOW
                                  ;  BIT 1 PRESCALER OUTPUT       LOW
                                  ;  BIT 0 PRESCALER OUTPUT       LOW

                 LD    P2M,#%80   ;PORT 2 MODE REGISTER
                                  ;  BIT 7 ZERO-X         INPUT
                                  ;  BIT 6 PIEZO          OUTPUT
                                  ;  BIT 5 CHARGE ENABLE  OUTPUT
                                  ;  BIT 4 AD ARM/CONVERT OUTPUT

                                  ;  BIT 3 LED            OUTPUT
                                  ;  BIT 2 PRESCALER      OUTPUT
                                  ;  BIT 1 PRESCALER      OUTPUT
                                  ;  BIT 0 PRESCALER      OUTPUT
                 LD    P2M_IMAGE,#%80

                 LD    PRE0,#%ED   ;MODULO 59 FOR T0 TIMER ENABLED.
                 LD    T0,#237     ;ALLOWS FOR OVERFLOW AT 1/32 SEC.
                 LD    TMR,#03H    ;ENABLE T0

                 CLR   IMR         ;KILL UNEXPECTED INT ENABLES (T0=POLLED!)
                 AND   IRQ,#%C0    ;KILL ANY PENDING INTERRUPTS
                 EI                ;GLOBALLY ENABLE INTERRUPT CAPABILITY
                 DI                ;NOW DISABLE, PGM INITIALIZES INTS LATER

                 CALL  DELAY_2     ;TO SYNC

INIT_MSLOPE:     LD   MSLOPE_MSB,#^HB 999
                 LD   MSLOPE_LSB,#^LB 999

SYSTEM_REGS_OK:
                 RET

                       
PUT_ATTENUATOR:
                  AND  P2M_IMAGE,#11111000B
                  OR   P2M_IMAGE,ATTENUATOR
                  AND  P2,#11111000B
                  LD   P2M,P2M_IMAGE
                  AND  P2,#11111000B
                  RET


SET_MAXIMUM_ATTENUATOR:
                  CLR    ATTENUATOR
                  CALL   PUT_ATTENUATOR
                  RET

;******************** BATTERY REMOVAL TEST SUBROUTINE ************************
;**      An out-of-range high a/d value means the battery is gone           **
;*****************************************************************************

CHECK_FOR_BATTERY_REMOVAL
                 TM  USER_FLAGS,#IGNORE_REMOVAL
                 JR  NE,BATTERY_IS_INSERTED

                 TM  AD_MSB,#10H
                 JR  EQ,BATTERY_IS_INSERTED

                 TM USER_FLAGS,#CURRENT_STATE  ;IS CURRENT ALREADY OFF
                 JP EQ,SOFT_START  ;YES AND A/D IS OVERRANGE, BATTERY IS GONE

                            ;FORCE MAX ATTENUATOR
                            ;TURN OFF CURRENT
                            ;TAKE A/D READING
                            ;IF A/D > A/D_OVERANGE THEN JUMP TO SOFT_START
                            ;ELSE PROBLEM MODE

                 CALL  CURRENT_OFF
                 CALL  SET_MAXIMUM_ATTENUATOR
                 CALL  DELAY_5
                 CALL  GET_AD

                 TM  AD_MSB,#10H
                 JP  NE,SOFT_START

                 LD   PROBLEM_MODE,#OVERRANGE_W_I_ON_PROBLEM
                 JP   PROBLEM

BATTERY_IS_INSERTED
                 RET

                      
;************************* FACTORY TEST ROUTINE ******************************
;** This can be deleted to save ROM space, or modified to suit design needs ** 
;*****************************************************************************

TEST_MODE:      LD     SPL,#80H
                CLEAR  TIME
                LD     P3M,#03H                ;P2 IS PUSH-PULL
                LD     P2M,#10111111B          ;P2 BIT 6 PIEZO IS OUTPUT

PHASE_1_LOOP:
                OR     P2,#01000000B           ;P2 BIT 6 PIEZO OUTPUT HIGH
                CALL   PIEZO_DELAY             ;WAIT

                add    TIME_LSB,#01
                adc    TIME_MSB,#00

                AND    P2,#10111111B           ;P2 BIT 6 PIEZO OUTPUT LOW
                CALL   PIEZO_DELAY             ;WAIT

                TM     P2,#00001000B           ;LOOP TILL TEST MODE INPUT
                JR     Z,PHASE_1_LOOP          ;  LINE RELEASED.

                TEST2  TIME, ULT, 8000, POWER_UP1 ;MUST HAVE BEEN IN PIEZO_TEST
                                               ;  FOR AT LEAST 2 SECONDS.
                DELAY  100

                CALL  INIT_SYSTEM_REGS ;INITALIZE SYSTEM REGISTERS

                CALL  CURRENT_ON

                LD    MODE,#'T'

                LD    TEST_MODE_COUNTER,#30  ;LOOP UP TO 30 TIMES, ABOUT 1 MINUTE

PHASE_2_LOOP:   LD    TEST_MODE_COUNTER,#4
                XMIT_CHAR MODE


;*** SEND ATTENUATION 3 A/D...
PRESCALER_TEST: LD   ATTENUATOR,#3
                CALL PUT_ATTENUATION_AND_REPORT

;*** SEND ATTENUATION 5 A/D...
                LD   ATTENUATOR,#5
                CALL PUT_ATTENUATION_AND_REPORT

;*** SEND ATTENUATION 6 A/D...
                LD   ATTENUATOR,#6
                CALL PUT_ATTENUATION_AND_REPORT

                DEC  TEST_MODE_COUNTER
                JR   NE, PHASE_2_LOOP

                LD  PROBLEM_MODE,#INADVERTENT_TEST_MODE
                JP  PROBLEM

PUT_ATTENUATION_AND_REPORT:
                CALL PUT_ATTENUATOR
                DELAY 750
                CALL GET_AD
                XMIT_CHAR MODE
                ADD  ATTENUATOR,#'0'
                XMIT_CHAR ATTENUATOR
                XMIT_NUMERIC AD
                CALL SEND_CR_LF
                RET

                .BYTE   %FF

                .END
